#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
A runeberg.org work represents a digitised book.

This most often corresponds to an edition in e.g. Wikidata terminology.

@TODO: load some key metadata as attributes, e.g. title and (co)author
    (which requires a lookup of the key so parsing of a.lst)
"""
import os
from collections import Counter, OrderedDict
from itertools import takewhile
from shlex import quote
from shutil import which  # used for djvu conversion
from subprocess import DEVNULL, run  # used for djvu conversion

from tqdm import tqdm

from runeberg.article import Article
from runeberg.download import DATA_DIR, IMG_DIR, SITE, UNZIP_SUBDIR
from runeberg.lst_file import LstFile
from runeberg.page import Page
from runeberg.page_range import PageRange
from runeberg.person import Person

IMAGE_TYPES = ('.tif', '.jpg')


class Work(object):
    """An object representing a runeberg.org work."""

    # @TODO: consider adding more params
    def __init__(self, uid):
        """
        Initialise a Work.

        @param uid: unique work identifier on runeberg.org
        """
        self.uid = uid  # work identifier on runeberg.org
        self.pages = PageRange()  # the included pages
        self.articles = OrderedDict()  # the included articles
        self.metadata = {}  # a verbatim copy of the runeberg.org metadata
        self.identifiers = {}  # identifiers for the work in external sources
        self.image_sources = {}  # sources for non-runeberg.org originated scan
        self.people = {}  # people involved with the work
        self.djvu = None  # djvu file of the whole work
        self.image_type = None  # the default file extension of the image files

        # @TODO: something recording the date when runeberg was downloaded

    @staticmethod
    def from_files(uid, base_path=None, known_people=None, img_dir=None):
        """
        Create an Work from the downloaded, unzipped files.

        @param base_path: path to the directory containing the unzipped files.
            Overrides the default {cwd}/DATA_DIR/{uid}/UNZIP_SUBDIR
        @param known_people: dict of Person objects from which author, editor
            and translator roles are matched.
        @param img_dir: the name given to the image sub-directory. Defaults to
            IMG_DIR.
        """
        work = Work(uid)
        base_path = base_path or os.path.join(
            os.getcwd(), DATA_DIR, uid, UNZIP_SUBDIR)
        img_dir = img_dir or IMG_DIR
        work.determine_image_file_type(base_path, img_dir)
        chapters = work.load_pages(base_path)
        work.load_articles(base_path, chapters)
        work.load_metadata(base_path)
        work.identifiers = work.parse_marc()
        work.image_sources = work.parse_image_source()
        work.parse_people(known_people)
        return work

    @property
    def text(self):
        """Return all the text of the work."""
        return self.pages.text

    def runeberg_url(self):
        """Return the base url on runeberg.org."""
        return '{0}/{1}/'.format(SITE, self.uid)

    def determine_image_file_type(self, base_path, img_dir):
        """
        Determine the file type of the images of the scanned pages.

        Will stop after encountering the first recognised image file type.

        @raises NoImagesError: if no files are encountered
        @raises UnrecognisedImageTypeError: if no recognised image file types
            are encountered.
        @param base_path: The path to the unzipped directory.
        @param img_dir: the image subdirectory
        """
        found_types = set()
        img_dir_path = os.path.join(base_path, img_dir)
        with os.scandir(img_dir_path) as it:
            for entry in it:
                if not entry.name.startswith('.') and entry.is_file():
                    ext = os.path.splitext(entry)[1]
                    if ext in IMAGE_TYPES:
                        self.image_type = ext
                        return
                    found_types.add(ext)
        if not found_types:
            raise NoImagesError(img_dir_path)
        raise UnrecognisedImageTypeError(found_types)

    def load_pages(self, base_path):
        """
        Populate self.pages.

        This is done using the Pages.lst file and the contents of the Pages
        and Images directories.

        @param base_path: The path to the unzipped directory.
        @return: Counter of all chapter names
        """
        pages = LstFile.from_file(os.path.join(base_path, 'Pages.lst'))
        whole_page_ok = LstFile.from_file(
            os.path.join(base_path, 'Pages', 'whole-page-ok.lst'))
        ok_pages = [i[0] for i in whole_page_ok.data]

        chapters = Counter()
        for uid, label in pages.data:
            page = Page.from_path(
                base_path, uid, self.image_type, label, uid in ok_pages)
            chapters.update(page.get_chapters())
            self.pages[uid] = page
        return chapters

    @staticmethod
    def setup_disambiguation_counter(chapters):
        """Create a Counter of non-unique chapter titles."""
        return Counter(
            [key for key, count in
             takewhile(lambda key_count: key_count[1] > 1,
                       chapters.most_common())])

    # @TODO: Implement require_unique_titles=False
    #   note that without requiring uniqueness self.articles must be a list...
    #   not necessarilly a problem though since we don't seem to use the index
    def load_articles(self, base_path, chapters, require_unique_titles=True,
                      reconcile_chapter_tags=True):
        """
        Populate self.articles.

        This is done using the Articles.lst file.

        This method allows the enforcement of two features which are not native
        to Runeberg.org.

        With `require_unique_titles=True` every article/chapter must have a
        unique name/title. When needed disambiguators will be added to article
        names. When disambiguators cannot be added an error is raised.

        With `reconcile_chapter_tags=True` every chapter tag, encountered when
        loading the pages, must correspond to an entry in Articles.lst. Entries
        in Articles.lst are still allowed to not have a counterpart in the
        chapter tags.

        @param base_path: The path to the unzipped directory.
        @param chapters: Counter of chapter tags encountered when loading the
            pages.
        @param require_unique_titles: Forces chapter titles to be unique.
        @param reconcile_chapter_tags: Forces reconciliation of article titles
            and encountered chapter tags.
        @return: self.articles
        """
        if require_unique_titles is False:
            raise NotImplementedError(
                'Sorry uniqueness is not optional for now')

        articles = LstFile.from_file(os.path.join(base_path, 'Articles.lst'))
        disambiguation_counter = Work.setup_disambiguation_counter(chapters)
        for html_name, title, page_range in articles.data:
            page_list = self.parse_range(page_range)
            if title in self.articles:
                raise DisambiguationError(title)
            if title in disambiguation_counter:
                article = Article(title, page_list, html_name,
                                  disambiguation_counter[title])
                page = article.first_page()
                page.rename_chapter(title, article.uid)
                disambiguation_counter[title] += 1
            else:
                article = Article(title, page_list, html_name)

            if not article.is_toc_header():
                if title not in article.first_page().get_chapters():
                    article.no_chapter_tag = True
                else:
                    chapters[title] -= 1
            self.articles[article.uid] = article

        # Do a reconciliation checks
        if reconcile_chapter_tags:
            if any(counter != 0 for counter in chapters.values()):
                raise ReconciliationError(chapters, self.articles)

        return self.articles

    # @TODO: move to staticmethod in article.py?
    def parse_range(self, page_range):
        """Parse the range of pages provided in Articles.lst."""
        page_list = []
        for entry in page_range.split(' '):
            if not entry.strip():
                continue
            if '-' in entry:
                first, _, last = entry.partition('-')
                page_list.append(self.pages.get_range(first, last))
            else:
                page_list.append(self.pages.get(entry))
        return page_list

    @staticmethod
    def read_metadata(base_path):
        """
        Read the metadata file and returns the lines.

        @param base_path: The path to the unzipped directory.
        @return: list of lines (incl. comments)
        """
        file_path = os.path.join(base_path, 'Metadata')
        with open(file_path) as f:
            return f.read().split('\n')  # unlike readlines() this trims the \n

    # @TODO: consider triggering parse_marc here
    def load_metadata(self, base_path):
        """
        Load the metadata, overwriting any previous value of that attribute.

        All the metadata is loaded into the metadata attribute, additionally
        some metadata fields are further processed.

        @param base_path: The path to the unzipped directory.
        @return: dict of metadata entries
        """
        metadata = {}
        data = Work.read_metadata(base_path)

        for line in data:
            if not line or line.startswith('#'):
                continue
            field, _, value = line.partition(':')
            if field in metadata.keys():
                raise NotImplementedError('Need to handle non-unique metadata')
            if not _:
                raise ValueError(
                    'The following metadata line is invalid: {}'.format(line))
            metadata[field] = value.lstrip()

        # sanity checks
        if metadata.get('CHARSET') != 'utf-8':
            raise NotImplementedError(
                'Need to handle handle non-utf8 works. Found {0}'.format(
                    metadata.get('CHARSET')))
        if metadata.get('TITLEKEY') != self.uid:
            raise ValueError(  # @TODO: appropriate Exception type?
                'Metadata work identifier does not match the provided '
                'identifier (provided: "{0}", found: "{1}")'.format(
                    self.uid, metadata.get('TITLEKEY')))
        self.metadata = metadata
        return metadata

    def parse_marc(self):
        """
        Parse the multivalued MARC metadata entry.

        @return dict of identifiers
        """
        data = self.metadata.get('MARC')
        return Work.parse_multivalued_mappings(data, 'identifiers')

    def parse_image_source(self):
        """
        Parse the multivalued IMAGE_SOURCE metadata entry.

        @return dict of image sources
        """
        data = self.metadata.get('IMAGE_SOURCE')
        return Work.parse_multivalued_mappings(data, 'image sources')

    @staticmethod
    def parse_multivalued_mappings(data, label):
        """
        Parse the multivalued mappings.

        Values are separated by a single space. Mapping pairs use a colon.

        @param data: string holding the multivalued mapping data to parse
        @param label: string describing the data for use in errors
        @return dict of mappings
        """
        parsed_data = {}
        if data:
            for entry in data.split(' '):
                field, _, value = entry.partition(':')
                if field in parsed_data.keys():
                    raise NotImplementedError(
                        'Need to handle non-unique {0}'.format(label))
                parsed_data[field] = value
        return parsed_data

    def parse_people(self, known_people=None):
        """
        Look up persons mentioned in metadata in the author file.

        If not provided the author file will be downloaded.
        """
        if not known_people:
            known_people = Person.load_people(DATA_DIR)
        people_fields = ('author', 'coauthor', 'translator')
        for field in people_fields:
            key = '{0}KEY'.format(field.upper())
            if self.metadata.get(key):
                uids = self.metadata.get(key)
                self.people[field] = [known_people[uid]
                                      for uid in uids.split(' ')]

    def to_djvu(self, silent=True, force=False):
        """
        Convert all images to a single DjVu file.

        Assumes only black-white pages and requires DjVuLibre.

        @param silent: silence the very chatty stderr output
        @type silent: Bool
        @param force: force re-generating the djvu, discarding any pre-existing
            file
        @type force: Bool
        @return: path to djvu file
        """
        base_path = os.path.dirname(self.pages.first().image)
        book_djvu = os.path.join(base_path, 'book.djvu')
        if not force:
            if self.djvu is not None:
                # checks if already converted
                # @TODO: notify the user somehow
                return self.djvu
            if os.path.isfile(book_djvu):
                # check if converted file already exists
                # @TODO: notify the user somehow
                self.djvu = book_djvu
                for i, page in enumerate(self.pages.values(), 1):
                    page.image_no = i
                return self.djvu
        if not Work.can_djvu():
            # @TODO: should find a more suitable exception
            raise Exception('Need DjVuLibre to convert to djvu')

        tmp_djvu = 'tmp.djvu'
        stderr = DEVNULL if silent else None
        for i, page in tqdm(enumerate(self.pages.values(), 1),
                            total=len(self.pages)):
            in_file = quote(page.image)

            # @TODO: can try "capture_output=True" if python is bumped to 3.7
            # use page file type in case original image has been replaced
            if page.image_file_type == '.tif':
                # cjb2 to convert single tif to djvu
                run(['cjb2', '-clean', in_file, tmp_djvu],
                    check=True, stderr=stderr)
            elif page.image_file_type == '.jpg':
                run(['c44', '-crcbfull', in_file, tmp_djvu],
                    check=True, stderr=stderr)
            else:
                raise NotImplementedError(
                    'At the moment only .tif and .jpg images can be converted '
                    'to DjVu')

            if i == 1:
                # djvm to create multi-page file from first page
                run(['djvm', '-c', book_djvu, tmp_djvu],
                    check=True, stderr=stderr)
            else:
                # djvm to insert page at end of multi-page file
                run(['djvm', '-i', book_djvu, tmp_djvu],
                    check=True, stderr=stderr)
            # remove is kept inside the loop to ensure cjb2 failure doesn't
            # result in a duplicated page
            os.remove(tmp_djvu)
            page.image_no = i
        self.djvu = book_djvu
        return book_djvu

    # @TODO: possibly move to utils or outside Work class
    @staticmethod
    def can_djvu():
        """
        Check if DjVu files can be created.

        Check whether DjVuLibre is installed, on PATH and marked as
        executable.
        """
        return (which('djvm') is not None and which('cjb2') is not None)


class NoImagesError(Exception):
    """Error when no images were found for the pages of the work."""

    def __init__(self, img_dir_path):
        """Initialise a NoImagesError."""
        msg = (
            'Could not detect any images for this work. Are you sure you '
            'provided the correct image subdirectory ({}) and that this is '
            'not a volumed work?'.format(img_dir_path))
        super().__init__(msg)


class UnrecognisedImageTypeError(Exception):
    """Error in trying to determine the image file type."""

    def __init__(self, encountered):
        """Initialise a UnrecognisedImageTypeError."""
        encountered = sorted(list(encountered))  # make deterministic
        msg = (
            'The encountered images do not seem to be of a recognised image '
            'file type.\n'
            'Recognised image file types are: {0}\n'
            'Encountered file types: {1}'.format(
                ', '.join(IMAGE_TYPES), ', '.join(encountered)))
        super().__init__(msg)


class DisambiguationError(Exception):
    """Error in disambiguating non-unique article titles."""

    def __init__(self, title):
        """Initialise a DisambiguationError."""
        msg = (
            'Encountered a duplicate article name ("{0}") which could '
            'not be automatically disambiguated. This is likely due '
            'to an error in the Articles.lst file which must be '
            'manually fixed. Encountered duplicates:\n'.format(
                '\n'.join(title)))
        super().__init__(msg)


class ReconciliationError(Exception):
    """Error in matching on-page chapter tags to articles in Articles.lst."""

    def __init__(self, chapters, articles):
        """Initialise a ReconciliationError."""
        self.unclaimed_chapters = ['{0}: {1} time(s)'.format(key, count)
                                   for key, count in chapters.most_common()
                                   if count != 0]
        self.no_tag_articles = ['{0}: page(s) {1}'.format(article.clean_title,
                                                          article.pages[0])
                                for article in articles.values()
                                if article.no_chapter_tag]

        msg = (
            'Encountered article names in the ocr:ed text which could not '
            'be found in Articles.lst. This is most often caused by typos '
            'or slight variations. These need to be manually reconciled '
            'before going forward. Note that not all articles need a '
            'corresponding chapter tag, but all chapter tags should '
            'correspond to an article.\n\n'
            '==Unclaimed chapter tags==\n'
            '{0}\n\n'
            '==Articles without chapter tags==\n'
            '{1}\n'.format('\n'.join(self.unclaimed_chapters),
                           '\n'.join(self.no_tag_articles)))
        super().__init__(msg)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
A runeberg.org person from the author file.

These people are normally authors, translators or co-authors/editors.

The columns in the author file are:
* birth,
* death,
* surname,
* first_name,
* nationalities,
* notes,
* uid.
"""
from datetime import datetime

import runeberg.download as downloader
from runeberg.lst_file import LstFile


class Person(object):
    """An object representing a runeberg.org person."""

    def __init__(self, uid, last_name, first_name='', birth_year=None,
                 death_year=None, nationalities=None, notes=None):
        """
        Initialise a Person.

        @param uid: unique person identifier on runeberg.org
        @param last_name: surname or primary sorting name people before circa
            year 1500 are listed by first name
        @param first_name: firstname or secondary sorting name,
        @param birth_year: year of birth, four digits or blank
        @param death_year: year of death, four digits or blank
        @param nationalities: list of two-letter ISO nationality codes or ??
            for unknown
        @param notes: notes, including:
            * alternative names e.g. before/after marriage
            * profession, e.g. writer, journalist,
            * pseudonyms, e.g. (pseud. Maria Lang),
            * signatures, e.g. (sign. S.L-d),
            * sources, e.g. (source: SVFF).
        """
        self.uid = uid
        self.last_name = last_name
        self.first_name = first_name
        self.birth_year = birth_year or None
        self.death_year = death_year or None
        self.nationalities = nationalities or []
        self.notes = notes

        try:
            self.birth_year = int(birth_year)
        except (TypeError, ValueError):
            pass
        try:
            self.death_year = int(death_year)
        except (TypeError, ValueError):
            pass

    def name(self):
        """Return the composite name of the person."""
        return '{0} {1}'.format(self.first_name, self.last_name).strip()

    def is_pd(self, years=70):
        """
        If copyrights by this person have expired.

        Note that the copyright expires on the first day of the following year.

        @params year: years after death at which copyright expires. Defaults to
            70.
        """
        print(datetime.now().year)
        if self.death_year and (datetime.now().year - self.death_year) > years:
            return True
        return False

    @classmethod
    def load_people(cls, data_dir=None):
        """Download and parse runeberg.org author file."""
        file_path = downloader.download_author_file(data_dir)
        return cls.people_from_file(file_path)

    # @TODO: Notes should be parsed further
    @classmethod
    def people_from_file(cls, file_path):
        """Parse runeberg.org author file."""
        lst_file = LstFile.from_file(file_path)
        people = {}
        for data in lst_file.data:
            birth, death, surname, first_name, nationalities, notes, uid = data
            people[uid] = cls(uid, surname, first_name, birth, death,
                              nationalities.split(','), notes)
        return people



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
A runeberg.org article is a chapter, article or poem.

Per http://runeberg.org/admin/19990511.html#Indexing_Articles:
'Project Runeberg's text editions present each chapter of a novel as an HTML
file of its own. Poetry book are made up of poems, while journals and
dictionaries have articles rather than chapters. The word article has been
chosen as the common term for the parts of any electronic edition.'

An article might span several pages, and one page can contain several articles.
The sequence of pages of an article need not be continuous as e.g.
advertisement might interrupt it.

Note that some entries in Articles.lst are only used as Table of Content
headers, as such they should not be treated as proper articles.
"""
# @TODO: Handle sub-articles. See e.g. "register" http://runeberg.org/aoshbok/
from bs4 import BeautifulSoup

from runeberg.page_range import PageRange


class Article(object):
    """An object representing a runeberg.org article."""

    # quick sketch
    def __init__(self, title, pages=None, html_name=None, disambig=None):
        """
        Initialise an Article.

        @param title: the article title
        @param pages: list of Pages or PageRanges which make up the article
        @param html_name: html file name used in runeberg.org
        @param disambig: disambiguator for none work-unique title
        """
        self.title = title  # the title, also used in headings on page
        self.pages = pages or []  # list of Pages or PageRanges
        self.html_name = html_name  # html file name used in runeberg.org
        self.disambig = disambig  # disambiguator for none work-unique title
        self.no_chapter_tag = None  # there is no matching chapter tag

    def first_page(self):
        """Return the first Page in the Article."""
        page = self.pages[0]
        if isinstance(page, PageRange):
            return page.first()
        return page

    def last_page(self):
        """Return the last Page in the Article."""
        page = self.pages[-1]
        if isinstance(page, PageRange):
            return page.last()
        return page

    def has_chapter_tag(self):
        """First page in the article has a matching chapter tag."""
        if self.is_toc_header() or self.no_chapter_tag:
            return False
        return True

    def is_toc_header(self):
        """
        Entry should only appear in the ToC.

        Some 'articles' are merely placeholders meant to inject an extra
        heading in a Table of Contents. E.g. 'B' indicating a change of letter
        in an alphabetic list of articles.
        """
        return not self.pages

    @property
    def clean_title(self):
        """
        Clean any html tags from the title.

        For Table of Contents purposes the name may include either empty
        <a> tags or various simpler formatting.
        """
        if not hasattr(self, '_clean_title'):
            soup = BeautifulSoup(self.title, features='html5lib')
            text = soup.get_text()
            self._clean_title = text.strip()
        return self._clean_title

    @property
    def text(self):
        """Return all the text of the article."""
        return '\n'.join([page.text for page in self.pages])

    @property
    def uid(self):
        """Work-unique title/id for the article."""
        if self.disambig:
            return '{0} ({1})'.format(self.clean_title, self.disambig)
        else:
            return self.clean_title

    def __eq__(self, other):
        """Implement equality comparison."""
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Entry point for navigating the works at Runeberg.org."""
import argparse
from collections import namedtuple

import runeberg.download as downloader
from runeberg.lst_file import LstFile

DEFAULT_PER_PAGE = 25
all_authors = None
all_works = None


def load_authors(in_use_authors):
    """
    Load the authors file.

    Loads all authors but only returns those which are used in works.
    @param in_use_authors: set of all author and coauthor uids encountered in
        works.
    """
    global all_authors
    if not all_authors:
        stream = downloader.download_author_file(save=False)
        # @TODO replace by runeberg.Person once #3 gets implemented
        Author = namedtuple('Author', 'birth death surname first_name '
                                      'nationalities notes uid')
        lst_authors = LstFile.from_stream(stream, func=Author)
        all_authors = {author.uid: author
                       for author in lst_authors.data
                       if author.uid in in_use_authors}


def load_works():
    """
    Load the works file.

    @return: set of all author and coauthor uids.
    """
    global all_works
    if not all_works:
        stream = downloader.download_works_file(save=False)
        Work = namedtuple('Work', 'title uid author_uids language year_start '
                          'year_end coauthor_uids translator_uids '
                          'original_language')
        lst_works = LstFile.from_stream(stream, func=Work)
        all_works = {work.uid: work for work in lst_works.data}

        in_use_authors = set()
        for work in lst_works.data:
            in_use_authors.update(work.author_uids.split())
            in_use_authors.update(work.coauthor_uids.split())
    return in_use_authors


# @TODO: year based filter
def filtered_work_generator(author=None, language=None, uid=None, **kargs):
    """
    Generate works matching the provided filter.

    @param author: author to filter on
    @param language: language to filter on
    @yield work
    """
    for work in all_works.values():
        if uid and uid != work.uid:
            continue
        if (author and author not in work.author_uids.split()
                and author not in work.coauthor_uids.split()):
            continue
        if language and language.lower() not in work.language.split():
            continue
        yield work


# @TODO: year based filters
def filtered_author_generator(nationality=None, uid=None, **kargs):
    """
    Generate authors matching the provided filter.

    @param nationality: nationality to filter on
    @yield work
    """
    for author in all_authors.values():
        if uid and uid != author.uid:
            continue
        if (nationality and
                nationality.lower() not in author.nationalities.split()):
            continue
        yield author


def year_range(y_start, y_end):
    """Format a year range."""
    year = y_start or ''
    if y_end:
        year += '–' + y_end
    return year


def author_as_string(author, short=False):
    """
    Format an author.

    Uses the format: first_name last_name (year_b-year_d) [nationalities]

    @param author: Author to format
    @param short: Whether to just return the name
    """
    name = '{0} {1}'.format(author.first_name, author.surname).strip()
    if short:
        return name
    year = year_range(author.birth, author.death)
    nat = ', '.join(author.nationalities.split())
    return '{name} ({year}) [{nat}]'.format(
        name=name, year=year or '?', nat=nat or '?')


def work_as_string(work):
    """
    Format a work.

    Uses the format: title (year) by author_1, author_2 and author_3 [langs]

    If the work has no authors it falls back on coauthors.
    """
    year = year_range(work.year_start, work.year_end)
    author_uids = work.author_uids.split() or work.coauthor_uids.split()
    authors = [author_as_string(all_authors[auth_uid], short=True)
               for auth_uid in author_uids]
    author_strings = ''
    if len(authors) == 1:
        author_strings = authors[0]
    elif len(authors) > 1:
        author_strings = '{0} and {1}'.format(
            ', '.join(authors[:-1]), authors[-1])

    return '{title} ({year}) by {authors} [{lang}]'.format(
        title=work.title, year=year or '?', authors=author_strings or '?',
        lang=work.language or '?')


def display_works(filters, per_page):
    """
    Display a filtered list of works for the user to choose from.

    @param filters: author and work filters to pass on
    @param per_page: results to output per page.
    @return: chosen work uid
    """
    chosen_work = pager(filtered_work_generator, filters, work_as_string,
                        'download', per_page=per_page)
    print("You picked '{0}' [uid={1}], downloading…".format(
        chosen_work.title, chosen_work.uid))
    return chosen_work.uid


def display_authors(filters, per_page):
    """
    Display a filtered list of authors for the user to choose works from.

    Upon choosing one the list of works is displayed.

    @param filters: author and work filters to pass on
    @param per_page: results to output per page.
    @return: chosen work uid
    """
    chosen_author = pager(filtered_author_generator, filters, author_as_string,
                          'display their works', per_page=per_page)
    print('Displaying works by {0} [uid={1}]…'.format(
        author_as_string(chosen_author, short=True), chosen_author.uid))
    filters['author'] = chosen_author.uid
    filters.pop('uid', None)  # since this was an author_uid not work_uid
    return display_works(filters, per_page)


def pager(generator, filters, as_string, select_action, per_page):
    """
    Page through the results from a generator, prompting a choice.

    @param generator: a generator for the entries to list.
    @param filters: filter which is passed on to the filtered_work_generator.
    @param as_string: function to convert generator entries to strings.
    @param select_action: description of what choosing an entry will result in.
    @param per_page: results to output per page.
    """
    displayed = []
    i = 0  # in case generator is empty
    for i, entry in enumerate(generator(**filters), 1):
        displayed.append(entry)
        print('{0}. {1}'.format(i, as_string(entry)))
        if i % per_page == 0:
            choice = prompt_choice(len(displayed), select_action, per_page)
            if choice is None:
                continue
            return displayed[choice - 1]

    if i == 0:
        print('Got no hits!, Sorry!')
        exit(0)
    else:  # handle the remainder
        if i % per_page == 0:
            print('That is all there was!, Sorry!')
        choice = prompt_choice(len(displayed), select_action, per_page=0)
        return displayed[choice - 1]


def prompt_choice(length, select_action, per_page):
    """
    Prompt the user for a choice of entry, to continue or to quit.

    An invalid choice will repeat the prompt.

    @param length: the largest choosable value
    @param select_action: description of what choosing an entry will result in.
    @param per_page: number of results to offer next. Set to 0 to hide "next"
        option.
    """
    prompt = 'What do you want to do? [{0}] to {1}, {2}[Q]uit: '.format(
        '1' if length == 1 else '1–{length}',
        select_action,
        '[N]ext {per_page}, ' if per_page else '')
    while True:
        choice = input(prompt.format(length=length, per_page=per_page))
        try:
            int_choice = int(choice)
        except ValueError:
            int_choice = None

        if choice.lower() == 'n' and per_page:
            return None
        elif choice.lower() == 'q':
            exit(0)
        elif int_choice and (1 <= int_choice <= length):
            return int_choice
        else:
            print('Invalid choice. Try again!')


class UpdateFilters(argparse.Action):
    """Action whereby the value is stored in a predefined dictionary."""

    def __call__(self, parser, namespace, values, option_string=None):
        """Store the arg and value as a key-value pair in a 'filters' dict."""
        adict = getattr(namespace, 'filters')
        adict.update({self.dest: values})


def handle_args(argv=None):
    """
    Parse and handle command line arguments.

    @param argv: arguments to parse. Defaults to sys.argv[1:].
    @return: argparse.Namespace
    """
    parser = argparse.ArgumentParser(
        description='Navigate Runeberg.org works and select one to download.')
    parser.set_defaults(filters={})
    parser.add_argument('-a', '--list_authors', dest='display_entries',
                        action='store_const', const=display_authors,
                        default=display_works,
                        help=('switch to first presenting a list of authors, '
                              'selecting an author presents the list of their '
                              'works.'))
    parser.add_argument('-n', '--per_page', type=int, default=DEFAULT_PER_PAGE,
                        action='store', metavar='N',
                        help=('number of results to output per go. Defaults '
                              'to {}.'.format(DEFAULT_PER_PAGE)))
    parser.add_argument('--dir', action='store', metavar='PATH',
                        help=('path to directory where files should be '
                              'downloaded.'))
    parser.add_argument('--update', action='store_true',
                        help='forced update of any previously downloaded '
                             'files.')
    parser.add_argument('--dry', action='store_true',
                        help='simulated run without the final download.')

    # filters
    parser.add_argument('--lang', dest='language', action=UpdateFilters,
                        metavar='XX', default=argparse.SUPPRESS,
                        help=('filter for works in this language '
                              '(2 letter code)'))
    parser.add_argument('--nationality', action=UpdateFilters,
                        metavar='XX', default=argparse.SUPPRESS,
                        help=('filter for authors with this nationality, '
                              '(2 letter ISO code).'))
    parser.add_argument('--uid', action=UpdateFilters,
                        default=argparse.SUPPRESS,
                        help=('the author_uid to filter on, if the -a flag is '
                              'present, else the work_uid to download.'))
    return parser.parse_args(argv)


def main():
    """Run main process."""
    args = handle_args()
    in_use_authors = load_works()
    load_authors(in_use_authors)
    work_uid = args.display_entries(args.filters, args.per_page)
    if not args.dry:
        downloader.get_work(work_uid, data_dir=args.dir, update=args.update)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Handler for the .lst file format, a fairly simple text document.

See definition at http://runeberg.org/admin/19990511.html
"""
import os


class LstFile(object):
    """Handler for the Lst file format."""

    CMT_SYMBOL = '#'
    DELIMITER = '|'

    def __init__(self, func=None):
        """
        Construct a file object.

        @param func: a function to which each parsed line is passed. If non is
            provided each line is returned as a tuple
        """
        self.data = []
        self.comments = []
        self.name = ''
        self._func = func

    @staticmethod
    def from_file(file_path, func=None):
        """
        Create an LstFile from an .lst file object.

        @param func: a function to which each parsed line is passed. If none is
            provided each line is returned as a tuple.
        """
        lst = LstFile(func)
        with open(file_path) as f:
            lst.name = os.path.basename(f.name)
            for line in f:
                lst.parse_line(line)
        return lst

    @staticmethod
    def from_stream(stream, file_name=None, func=None):
        """
        Create an LstFile from an .lst file stream.

        @param func: a function to which each parsed line is passed. If none is
            provided each line is returned as a tuple.
        """
        lst = LstFile(func)
        if file_name:
            lst.name = file_name
        for line in stream.split('\n'):
            lst.parse_line(line)
        return lst

    def parse_line(self, line):
        """Parse a line in an .lst file separating data from comments."""
        line = line.rstrip('\n')
        if not line:
            # skip completely empty lines
            pass
        elif line.startswith(LstFile.CMT_SYMBOL):
            # skip completly empty comments
            if line[len(LstFile.CMT_SYMBOL):].strip():
                self.comments.append(
                    line[len(LstFile.CMT_SYMBOL):].lstrip())
        else:
            d = line.split(LstFile.DELIMITER)
            if self._func:
                d = self._func(*d)
            else:
                d = tuple(d)
            self.data.append(d)

    @property
    def columns(self):
        """Determine the number of columns in the data set."""
        try:
            return max(len(row) for row in self.data)
        except TypeError:
            if self._func:
                raise NotImplementedError(
                    '{0} does not specify a length'.format(
                        self._func.__name__))
            raise

    def __len__(self):
        """Set length of file to number of data entries."""
        return len(self.data)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Download a single book from Runeberg.org.

@TODO: handle volumed works:
    ocr zip contains no Pages folder;
    images zip may not exist but r.status_code is still 200;
    article.lst may contain a list of volumes (see e.g. kok) but unclear if
    that is standard (see e.g. nf).
@TODO: Create update prompt for whether to overwrite downloaded files or not
@TODO: Consider dropping tqdm:
    Less useful as runeberg doesn't provide file size in header, but still
    indicates that something is happening (good for large files).
"""
import os
import shutil
import warnings
import zipfile

import requests
from tqdm import tqdm

SITE = 'http://runeberg.org'
DATA_DIR = 'downloaded_data'
UNZIP_SUBDIR = 'tmp'
IMG_DIR = 'Images'


def exists_on_runeberg(uid):
    """
    Check if the work exists on Runeberg.org.

    This does not check if the provided title is that of a work or the landing
    page for a multi-volumed work.

    @param uid: the runeberg.org work identifier to look for
    @return: bool
    """
    url = '{0}/{1}/'.format(SITE, uid)
    r = requests.get(url)
    return r.status_code == 200


def is_empty_file(file_path):
    """
    Check if the provided file path is that of an empty (0 byte) file.

    @param file_path: the file path to check
    @return: bool
    """
    return os.stat(file_path).st_size == 0


def download_author_file(data_dir=None, save=True):
    """
    Download the runeberg.org author file to a.lst and convert it to utf-8.

    @param data_dir: the directory to which the data should be downloaded.
        Defaults to DATA_DIR.
    @param save: whether the file should be saved or the stream returned
    @return: path to downloaded file if save, else the downloaded data
    """
    return download_lst_file('a.lst', data_dir, save)


def download_works_file(data_dir=None, save=True):
    """
    Download the runeberg.org works file to t.lst and convert it to utf-8.

    @param data_dir: the directory to which the data should be downloaded.
        Defaults to DATA_DIR.
    @param save: whether the file should be saved or the stream returned
    @return: path to downloaded file if save, else the downloaded data
    """
    return download_lst_file('t.lst', data_dir, save)


def download_lst_file(file_name, data_dir=None, save=True):
    """
    Download a runeberg.org .lst file and convert it from latin1 to utf-8.

    @param file_name: name of the file to download and to which it is saved.
    @param data_dir: the path of the directory to which the data should be
        saved once downloaded. Defaults to {cwd}/DATA_DIR.
    @param save: whether the file should be saved or the stream returned
    @return: path to downloaded file if save, else the downloaded data
    """
    url = '{0}/authors/{1}'.format(SITE, file_name)
    r = requests.get(url)
    r.encoding = 'latin1'

    if not save:
        return r.text

    data_dir = data_dir or os.path.join(os.getcwd(), DATA_DIR)
    output_file = os.path.join(data_dir, file_name)
    with open(output_file, 'w', encoding='utf-8') as f:
        f.write(r.text)
    return output_file


def get_work(uid, data_dir=None, sub_dir=None, img_dir='', update=None):
    """
    Download, unzip and normalise all of the files related to a work.

    @param uid: the runeberg.org work identifier for the work to download
    @param data_dir: the path of the directory to which any data should be
        downloaded (a work specific subdirectory will be created). Defaults
        to {cwd}/DATA_DIR.
    @param sub_dir: the name of the sub-directory of 'data_dir' to which the
        files should be unzipped. Any files already in this directory will be
        deleted. Defaults to UNZIP_SUBDIR.
    @param img_dir: the name given to the image sub-directory. Defaults to
        IMG_DIR. Set to None to keep the non-standardised name
    @param update: whether to overwrite already downloaded files. Set to None
        to trigger a prompt.
    """
    data_dir = data_dir or os.path.join(os.getcwd(), DATA_DIR)
    sub_dir = sub_dir or UNZIP_SUBDIR
    img_dir = IMG_DIR if img_dir == '' else img_dir

    work_dir = os.path.join(data_dir, uid)
    files = download_work(uid, work_dir, update=update)
    unzip_dir = unzip_work(files, sub_dir)
    normalise_unzipped_files(unzip_dir, img_dir)

    # test for volumed work
    if not os.path.isdir(os.path.join(unzip_dir, 'Pages')):
        raise NotImplementedError(
            'Looks like a volumed work was encountered. These are not yet '
            'supported, you can most often find the individual volumes listed '
            'in the Articles.lst file.')

    # test for colour images
    if not is_empty_file(os.path.join(work_dir, '{}_colour.zip'.format(uid))):
        warnings.warn(
            'There seem to be colour images available for this work. To use '
            'these in the parser they must be manually included after '
            'initialisation. Note that DjVu conversion may not work with '
            'coloured images.',
            UserWarning)


def download_work(uid, work_dir, update=None):
    """
    Download all of the files related to a work.

    @param uid: the runeberg.org work identifier for the work to download
    @param work_dir: the path of the work specific sub-directory to which the
        data should be downloaded.
    @param update: whether to overwrite already downloaded files. Set to None
        to trigger a prompt.
    @return: list of paths to downloaded files
    """
    if not exists_on_runeberg(uid):
        raise NoSuchWorkError(uid)
    try:
        os.makedirs(work_dir, exist_ok=True)
    except OSError as e:
        raise DirCreationError(work_dir, e)

    urls = [
        ('ocr', '{0}/download.pl?mode=txtzip&work={1}'.format(SITE, uid)),
        ('images', '{0}/{1}.zip'.format(SITE, uid)),
        ('colour', '{0}/download.pl?mode=jpgzip&work={1}'.format(SITE, uid))
    ]

    files = []
    for label, url in urls:
        output_file = os.path.join(
            work_dir, '{0}_{1}.zip'.format(uid, label))

        # handle already downloaded files
        if not update and (os.path.isfile(output_file)
                           and not is_empty_file(output_file)):
            if update is None:
                raise NotImplementedError('Update prompt not implemented')
            else:
                files.append(output_file)
                continue

        r = requests.get(url, stream=True)
        chunk_size = 1024
        with open(output_file, 'wb') as handle:
            pbar = tqdm(desc=label, unit_scale=True, unit='B')
            for data in r.iter_content(chunk_size=chunk_size):
                if data:  # filter out keep-alive new chunks
                    pbar.update(len(data))
                    handle.write(data)
        files.append(output_file)
    return files


def unzip_work(files, sub_dir):
    """
    Unzip all of the downloaded files for a work to the given sub-directory.

    @param files: list of zip files to unpack
    @param sub_dir: the name of the sub-directory to which the files should be
        unzipped. Any files already in this directory will be deleted.
    @return: path to the directory in which the files were unzipped
    """
    work_dir = os.path.dirname(files[0])
    unzip_dir = os.path.join(work_dir, sub_dir)
    try:
        os.mkdir(unzip_dir)  # will raise FileExistsError if exists
    except FileExistsError:
        # overwrite old unzips to ensure newest data only
        shutil.rmtree(unzip_dir)
        os.mkdir(unzip_dir)

    for f in files:
        # non-existent files end up as 0 byte zips
        # colour images risk overwriting bw images
        if is_empty_file(f) or f.endswith('_colour.zip'):
            continue
        with zipfile.ZipFile(f, 'r') as zip_ref:
            zip_ref.extractall(unzip_dir)

    return unzip_dir


def normalise_unzipped_files(unzip_dir, img_dir):
    """
    Normalise the directory structure and file encoding of the unzipped files.

    The images are stored in a directory named after the work uid, this is
    renamed to ensure the same structure for all unzips.

    The metadata file is 'ISO-8859-1' encoded, this is changed to utf-8.

    @param unzip_dir: the path to the directory to which the files were
        unzipped.
    @param img_dir: the name given to the image directory. If None the
        non-standardised name is kept.
    """
    work_uid = os.path.basename(os.path.split(unzip_dir)[0])

    # normalise image directory
    if img_dir and os.path.isdir(os.path.join(unzip_dir, work_uid)):
        os.rename(
            os.path.join(unzip_dir, work_uid),
            os.path.join(unzip_dir, img_dir))

    # normalise metadata file encoding
    metadata_file = os.path.join(unzip_dir, 'Metadata')
    if os.path.isfile(metadata_file):
        re_encode_file(metadata_file, 'ISO-8859-1', 'utf-8')


def re_encode_file(input_file, from_codec, to_codec):
    """Change the encoding of a file."""
    output_file = '{}_tmp'.format(input_file)

    with open(input_file, encoding=from_codec) as f, \
            open(output_file, 'w', encoding=to_codec) as e:
        text = f.read()
        e.write(text)
    os.remove(input_file)
    os.rename(output_file, input_file)


class NoSuchWorkError(Exception):
    """Error raised when a work does not exist."""

    def __init__(self, uid):
        """Initialise the Error."""
        msg = 'The work "{0}" was not found on runeberg.org'.format(uid)
        super().__init__(msg)


class DirCreationError(Exception):
    """Error raised when directory creation fails."""

    def __init__(self, out_path, error):
        """Initialise the Error."""
        msg = 'Could not create the directory "{0}". Error: {1}'.format(
            out_path, error)
        super().__init__(msg)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""A runeberg.org page is a single digitized page of a work."""
import os
import warnings

from bs4 import BeautifulSoup

from runeberg.download import IMG_DIR


class Page(object):
    """An object representing a runeberg.org page."""

    BLANK_LABELS = ('blank', '(blank)')  # labels known to indicate blank page
    DEFAULT_BLANK = '<blank>'  # default label to use for blank pages

    def __init__(self, uid, label=None, image=None, text=None, proofread=None):
        """
        Initialise a Page.

        @param uid: the identifier of the page, ordinarily the zero-padded
            ordinal number of the page.
        @param label: the real printable label of the page, ordinarily the page
            number.
        @param image: path to the image of the page
        @param text: ocr:ed (and potentially proofread) text of the page
        @param proofread: whether the page has been fully proofread.
        """
        self.uid = uid  # unique identifier of the page (often a padded number)
        self.label = label or ''  # what the numbering on the actual page is
        self.image = image or ''  # path to the image
        self.image_no = None  # page in djvu file on which the image appears
        self.text = text or ''  # ocr:ed text of the page
        self.is_proofread = proofread  # bool or None in the case of blank page
        self.set_blank()

    def __str__(self):
        """Represent the Page as a string."""
        return self.uid

    @property
    def image_file_type(self):
        """Return the file extension of the image file."""
        if not self.image:
            return None
        return os.path.splitext(self.image)[1]

    def set_blank(self, default_label=None):
        """
        Set the blank proofread status based on page label.

        Also changes the page label to a default_label.

        @param default_label: the default label to use for blank pages.
        """
        default_label = default_label or self.DEFAULT_BLANK
        if self.label in self.BLANK_LABELS:
            self.is_proofread = None
            self.label = default_label

    def check_blank(self):
        """Sanity check blank pages and blank page candidates."""
        if self.is_proofread is None and self.text.strip() != '':
            warnings.warn(
                '{0} ({1}) was labelled "blank" but text is not empty. Either '
                'clear out the ocr text or set `is_proofread=False`'.format(
                    self.uid, self.label),
                UserWarning)
        elif self.is_proofread and self.text.strip() == '':
            warnings.warn(
                '{0} ({1}) was labelled as proofread but is blank. Either '
                're-add the ocr text or set `is_proofread=None`'.format(
                    self.uid, self.label),
                UserWarning)
        elif self.is_proofread is False and self.text.strip() == '':
            warnings.warn(
                '{0} ({1}) might be blank. Check the image and if truly blank '
                'set `is_proofread=None`'.format(
                    self.uid, self.label),
                UserWarning)

    @staticmethod
    def from_path(base_path, uid, image_type, label, whole_page_ok):
        """
        Create a page from the path to the unzipped files and a uid.

        @param base_path: the file path to the directory containing the
            unzipped files.
        @param uid: the identifier of the page, ordinarily the zero-padded
            ordinal number of the page.
        @param image_type: the expected file extension of the scanned image.
        @param label: the real printable label of the page, ordinarily the page
            number.
        @param whole_page_ok: whether the whole page has been proofread
        """
        img_path = os.path.join(
            base_path, IMG_DIR, '{0}{1}'.format(uid, image_type))
        if not os.path.isfile(img_path):
            raise ValueError(
                '{} was provided as the image for "{}" but no such image '
                'exists'.format(img_path, uid))
        text_path = os.path.join(base_path, 'Pages', '{}.txt'.format(uid))
        with open(text_path) as f:
            text = f.read()
        page = Page(uid, label, img_path, text, whole_page_ok)
        page.check_blank()
        return page

    # @TODO: might change to @property, and rename self.chapters
    def get_chapters(self):
        """Extract all chapter names on page."""
        if not hasattr(self, '_chapters'):
            soup = BeautifulSoup(self.text, features='html5lib')
            self._chapters = [chapter.get('name')
                              for chapter in soup.find_all('chapter')]
            if any(not chapter for chapter in self._chapters):
                raise ValueError(
                    'Encountered a blank/missing chapter name on page '
                    '{0} ({1})'.format(self.uid, self.label))
        return self._chapters

    def rename_chapter(self, old_name, new_name):
        """Rename the first occurrence of a chapter."""
        base_str = '<chapter name="{}">'
        self.text = self.text.replace(
            base_str.format(old_name),
            base_str.format(new_name),
            1)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""The initialization file for the runeberg library."""
from runeberg.article import Article
from runeberg.page import Page
from runeberg.page_range import PageRange
from runeberg.person import Person
from runeberg.work import Work

__all__ = ('Article', 'Page', 'PageRange', 'Person', 'Work')



# -*- coding: utf-8 -*-
import setuptools

with open("README.md", "r") as readme, open("CHANGES.md", "r") as changes:
    long_description = '{0}\n\n{1}'.format(readme.read(), changes.read())

setuptools.setup(
    name="runeberg",
    version="0.0.2",
    author="André Costa",
    author_email="lokal.profil@gmail.com",
    description="Library for working with works from Projekt Runeberg (Runeberg.org).",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lokal-profil/runeberg",
    packages=["runeberg"],
    include_package_data=True,
    install_requires=["beautifulsoup4", "html5lib", "requests", "tqdm"],
    keywords=['runeberg', 'libraries', 'literature'],
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
            "runeberg=runeberg.__main__:main",
        ]
    },
)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""A sequence of consecutive Pages."""
from collections import OrderedDict
from itertools import islice


class PageRange(OrderedDict):
    """A non interrupted sequence of consecutive Pages."""

    def index(self, *args):
        """
        Find the index, or indexes, of the provided key(s).

        @return: list of indexes if multiple keys are provided otherwise a
            single index.
        """
        key_list = list(self.keys())
        if len(args) == 0:
            raise ValueError('At least one index required')
        elif len(args) == 1:
            return key_list.index(args[0])
        else:
            return [key_list.index(arg) for arg in args]

    def get_range(self, first, last, inclusive=True):
        """Create a sub-PageRange from between the provided values."""
        # the below should deal with 0 < last < first
        first_i, last_i = self.index(first, last)
        return self.slice(first_i, last_i + (1 * inclusive))

    def first(self):
        """Return the first page in the PageRange."""
        return next(iter(self.values()), None)

    def last(self):
        """Return the last page in the PageRange."""
        return next(reversed(self.values()), None)

    def __str__(self):
        """Represent the PageRange as a string."""
        # first-last
        # deal with zero pages?
        key_list = list(self.keys())
        if len(self) == 0:
            return ''
        if len(self) == 1:
            return key_list[0]
        else:
            return '{}-{}'.format(key_list[0], key_list[-1])

    @property
    def text(self):
        """Return the text from all the pages in the range."""
        return '\n'.join([page.text for page in self.values()])

    # not implemented in __getitem__ to avoid it mixing keys and indexes
    # at the cost of a bulkier syntax
    def slice(self, start=None, stop=None, step=None):
        """Add slicing capabilities missing in OrderedDict."""
        return PageRange(islice(self.items(), start, stop, step))

    # def __getitem__(self, k):
    #     """Add slicing to __getitem__ of OrderedDict."""
    #     if not isinstance(k, slice):
    #         return OrderedDict.__getitem__(self, k)
    #     return PageRange(islice(self.items(), k.start, k.stop))
